import unittest

from powerset_generator import subsets


class TestPowerset(unittest.TestCase):
    three_list: list[int] = [1, 2, 3]

    three_repetition: list[int] = [1, 2, 3, 1, 2, 2, 3, 3, 3]

    empty_set: set[int] = set()

    power_three: list[set[int]] = [
        empty_set,
        {1},
        {2},
        {3},
        {1, 2},
        {1, 3},
        {2, 3},
        {1, 2, 3},
    ]

    def test_powerset(self) -> None:
        result: list[set[int]] = [s for s in subsets(self.three_list)]

        # AssertCountEqual requires elements to be hashable, which they
        # aren't in our case. So we do it the slow way
        self.assertEqual(len(result), 2 ** len(self.three_list))
        self.assertTrue(
            all(result.count(i) == self.power_three.count(i) for i in result)
        )

    def test_powerset_repeated(self) -> None:
        result: list[set[int]] = [s for s in subsets(self.three_repetition)]

        # AssertCountEqual requires elements to be hashable, which they
        # aren't in our case. So we do it the slow way
        self.assertEqual(len(result), 2 ** len(self.three_list))
        self.assertTrue(
            all(result.count(i) == self.power_three.count(i) for i in result)
        )

    def test_cardinality(self) -> None:
        vectors: list[tuple[int, list[int]]] = [
            (1, []),
            (1024, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            (16, [1, 2, 3, 4, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4]),
            # ( "ℵ_1", ["the natural numbers"]),
        ]

        for v in vectors:
            count = 0
            for _ in subsets(v[1]):
                count += 1

            self.assertEqual(v[0], count)


if __name__ == "__main__":
    unittest.main()



from itertools import combinations
from collections import Counter
from collections.abc import Collection, Generator, Hashable
from typing import TypeVar

from . import __about__  # noqa

T = TypeVar("T", bound=Hashable)


def subsets(
    collection: Collection[T],
) -> Generator[set[T], None, None]:
    """Generates all the subsets of the powerset of collection.

    :param collection: The collection you want the subsets of

    Power sets
    ------------

    The power set of a collections is the set of all subsets of
    the collection.

    * The power set includes the empty set.

    * The power set includes a set of all members of the collections.
      That is, it is not limited to the proper subsets of the collection.

    * If the collection has *N* unique elements,
      the power set will have 2^N` elements.

    Usage notes
    -------------

    This this fuction generates sets irrespective of the type of its input.
    That is, given a list, it will yield sets.

    In the current version there is no enforcement of the type hinting
    that members of the input collection all be of the same type. But
    do not rely on this lax behavior.

    Example::

        from powerset_generator import subsets

        a_list: list[str] = ["one", "two", "three"]
        for element in subsets(s):
            print(element)

    produces::

        set()
        {'one'}
        {'three'}
        {'two'}
        {'one', 'three'}
        {'one', 'two'}
        {'three', 'two'}
        {'one', 'three', 'two'}

    Exceptions
    ----------

    powerset does not explicitly raise any exceptions on its own,
    but if its argument is not Collection[Hashable] some of the
    functions it calls internally are likely to raise a TypeError.
    """

    # We don't want duplicats, so
    s: set[T] = set(Counter(collection))

    # explicitly start at 0 to not forget that the empty set
    # is in the powerset
    for r in range(0, len(s) + 1):
        for result in combinations(s, r):
            # itertools.combinations spits out tuples
            yield set(result)



# There has to be a better way
import os
import sys
import tomllib

from powerset_generator import __about__

sys.path.insert(0, os.path.abspath("."))
sys.path.insert(0, os.path.abspath("../../src"))

# Pull general sphinx project info from pyproject.toml
with open("../../pyproject.toml", "rb") as f:
    toml = tomllib.load(f)

version = __about__.__version__

pyproject = toml["project"]

project = pyproject["name"]
release = version
author = ",".join([author["name"] for author in pyproject["authors"]])
copyright = f"2024 {author}"

rst_prolog = f"""
.. |project| replace:: **{project}**
.. |root| replace:: :mod:`powerset_generator`
"""


extensions: list[str] = [
    "sphinx.ext.autodoc",
    "sphinx_autodoc_typehints",
    "sphinx.ext.doctest",
]

# ## doctest setup
doctest_global_setup = """
from powerset_generator import subsets
"""


extensions.append("sphinx.ext.intersphinx")
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

extensions.append("sphinx.ext.mathjax")

templates_path = ["_templates"]
exclude_patterns: list[str] = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]
